﻿// SpeechBalloonHeadTracking - Using modified version of Marilena (http://www.libspark.org/wiki/mash/Marilena) created by Mario Klingemann / Quasimondo (http://www.quasimondo.com/archives/000687.php#000687)
// A demo performs real-time webcam head tracking and positions a speech balloon at the correct position and with the correct size
//  It also allows you to edit the text within this speech balloon in real-time.
// 
// released under MIT License
// http://www.opensource.org/licenses/mit-license.php
//
//@author		Devin Reimer
//@version		1.0.0
//@website 		http://blog.almostlogical.com, http://www.permaven.com


//Copyright (c) 2009 Devin Reimer

/*
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com
{
	import fl.controls.CheckBox;
	import fl.controls.TextArea;
	import flash.text.TextFieldAutoSize;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.TextEvent;
	import flash.geom.Matrix;
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.media.Video;
	import flash.display.Graphics;
	import flash.text.TextField;
	import flash.text.TextFormat;
	import flash.display.StageScaleMode;
	
	import jp.maaash.ObjectDetection.ObjectDetector;
	import jp.maaash.ObjectDetection.ObjectDetectorOptions;
	import jp.maaash.ObjectDetection.ObjectDetectorEvent;
	
	public class SpeechBalloonHeadTracking extends Sprite
	{	
		private static const MILLISECONDS_DELAY_BETWEEN_DETECTIONS:int = 300; //delay between head detections
		private static const DETECTION_HORIZ_MARGIN:int = 40; //the left and right margins which head tracking will not be done
		private static const DETECTION_VERT_MARGIN:int = 15; //the top and bottom margisn which head tracking will not be done
		private static const DETECTION_MIN_SIZE:int = 65; //minimum size of the head to be tracked
		
		private static const RE_DETECTION_MARGIN:int = 50; //the margin around a detected box which it will look on second try
		
		private var webcam:Webcam;
		private var mainVideo:Video;
		
		private var baseContainer:Sprite; //a container used to store all the page content
		
		private var webcamBitmap:Bitmap;
		private var webcamMatrix:Matrix;
		
		private var detector:ObjectDetector;
		private var detectionReady:Boolean = true;
		
		private var faceRectContainer:Sprite; //container that stores the drawing information for the head tracking rectangle
		
		private var currentHeadRect:Rectangle = new Rectangle(0, 0, 0, 0);
		
		private var lastDetectionStartTime:Number = -1;
		
		private var numberOfFailure:Number = 1; //number of times it has failed to find rectangle
		
		private var options:ObjectDetectorOptions;
		
		private var speechBalloon:SpeechBalloon; //the speech balloon where text will be displayed
		private var speechBalloonTF:TextField; //the text for the speech ballon
		
		private var speechTextArea:TextArea; //the text area for text input for the speech balloon
		private var trackingBoxCheckBox:CheckBox; //a checkbox used to enable and disable head tracking box
	
		public function SpeechBalloonHeadTracking()
		{
			stage.quality = "MEDIUM";
			stage.scaleMode = StageScaleMode.NO_SCALE;
			
			webcam = new Webcam();
			webcam.addEventListener(Event.COMPLETE, addVideo, false, 0, true);
		}
		private function addVideo(evt:Event)
		{
			webcamMatrix = new Matrix();
			webcam.removeEventListener(Event.COMPLETE, addVideo);	
			mainVideo = webcam.getVideo();
			webcamBitmap = new Bitmap();
			
			webcamMatrix.scale((mainVideo.scaleX),(mainVideo.scaleY)) //scale the transformation matrix to match the scale of the Video object on stage		
			webcamBitmap.bitmapData = new BitmapData(mainVideo.width, mainVideo.height,false); //create a bitmap object in memory that is the same size as the Video object
			
			initDetector();	
			lastDetectionStartTime = -1;
		}
		
		private function initDetector():void
		{
			options= new ObjectDetectorOptions();
			detector = new ObjectDetector();
			options.search_mode = ObjectDetectorOptions.SEARCH_MODE_NO_OVERLAP;
			//sets bounds increase performance
			options.min_size = DETECTION_MIN_SIZE;
			options.startx = DETECTION_HORIZ_MARGIN;
			options.starty = DETECTION_VERT_MARGIN;
			options.endx = mainVideo.width - DETECTION_HORIZ_MARGIN;
			options.endy = mainVideo.height - DETECTION_VERT_MARGIN;
			detector.options = options;
			detector.addEventListener(ObjectDetectorEvent.DETECTION_COMPLETE, detectionComplete, false, 0, true);
			/* no longer need because of the new modified Marilena class 
			detector.addEventListener(ObjectDetectorEvent.HAARCASCADES_LOAD_COMPLETE, detectorReady, false, 0, true);
			detector.loadHaarCascades("face.zip");	
			*/
			setup();
		}
	
		/* no longer need because of the new modified Marilena class
		private function detectorReady(evt:ObjectDetectorEvent):void 
		{
			detector.removeEventListener(ObjectDetectorEvent.HAARCASCADES_LOAD_COMPLETE,detectorReady);
			setup();
		}
		*/
		
		private function setup():void
		{
			faceRectContainer = new Sprite();
			
			baseContainer = new Sprite();
			baseContainer.x = (stage.stageWidth - mainVideo.width)/2;
			baseContainer.y = (stage.stageHeight - mainVideo.height)/2;
			
			baseContainer.addChild(mainVideo);
			baseContainer.addChild(faceRectContainer);
			
			//addChild(new FPS());
			
			speechBalloon = new SpeechBalloon();
			baseContainer.addChild(speechBalloon);
			
			addChild(baseContainer);
			
			speechTextArea = new TextArea();
			speechTextArea.focusEnabled = false;
			speechTextArea.width = mainVideo.width;
			speechTextArea.addEventListener(Event.CHANGE, updateSpeechBalloon, false, 0, true);
			speechTextArea.y = mainVideo.height;
			speechTextArea.setStyle("textFormat", new TextFormat(null, 15));
			speechTextArea.text = "Hi! I love head tracking.";
			baseContainer.addChild(speechTextArea);
			
			trackingBoxCheckBox = new CheckBox();
			trackingBoxCheckBox.y = speechTextArea.y + speechTextArea.height;
			trackingBoxCheckBox.textField.autoSize = TextFieldAutoSize.LEFT;
			trackingBoxCheckBox.setStyle("textFormat", new TextFormat(null, 14, 0xFFFFFF));
			trackingBoxCheckBox.label = "Enable/Disable Head Tracking Box Outline";
			baseContainer.addChild(trackingBoxCheckBox);
			
			speechBalloonTF = new TextField();
			speechBalloonTF.width = speechBalloon.width - 20;
			speechBalloonTF.height = speechBalloon.height -70;
			speechBalloonTF.multiline = true;
			speechBalloonTF.wordWrap = true;
			speechBalloonTF.defaultTextFormat = new TextFormat(null, 20);
			
			speechBalloonTF.x = 10;
			speechBalloonTF.y = 10;
			speechBalloon.addChild(speechBalloonTF);
			
			updateSpeechBalloon(null);
			
			addEventListener(Event.ENTER_FRAME, eachFrameTick, false, 0, true);
		}
		
		private function detectionComplete( evt:ObjectDetectorEvent ):void
		{
			var rectArea:Number = -1;
			var rectNum:int = -1;
			var r:Rectangle;
			if (evt.rects) //if tracking rectangles exist
			{			
				for (var rectCounter:int = 0; rectCounter < evt.rects.length; rectCounter++) //find largest rect
				{
					r = evt.rects[rectCounter];
					if (r.width * r.height > rectArea)
					{
						rectArea = r.width * r.height;
						rectNum = rectCounter;
					}
				}
				if (rectNum >= 0) //double check if rectangle that is suitable is found
				{
					r = evt.rects[rectNum];
					
					//this makes rect movement smoother
					currentHeadRect.x = (currentHeadRect.x + r.x) / 2;
					currentHeadRect.y = (currentHeadRect.y +r.y) / 2;
					currentHeadRect.width = (currentHeadRect.width + r.width) / 2;
					currentHeadRect.height = (currentHeadRect.height + r.height) / 2;
					
					numberOfFailure --; //removes 1 from the number of failures
					if (numberOfFailure < 0) { numberOfFailure = 0; } //number of failures can not be below 0
					renderInfoBasedOnHeadPosition(); //update position if required
				}
				else
				{
					numberOfFailure++; //add 1 to the number of failures
					if (numberOfFailure > 5) { numberOfFailure = 5; } //number of failures can not go above 5
				}
			}
			
			detectionReady = true;
		}
		
		private function eachFrameTick(evt:Event):void
		{
			findPattern(); //track head position
		}
		
		private function findPattern():void
		{
			//only if previous detection is complete and enough time has past do head tracking
			if (detectionReady && new Date().time - lastDetectionStartTime > MILLISECONDS_DELAY_BETWEEN_DETECTIONS)
			{
				detectionReady = false;
				webcamBitmap.bitmapData.draw(mainVideo, webcamMatrix, null, null, null, false);
				
				if (numberOfFailure <= 0) //this uses tighter bounds if no failures to increase speed
				{
					options.startx = currentHeadRect.x - RE_DETECTION_MARGIN;
					options.starty = currentHeadRect.y - RE_DETECTION_MARGIN;
					options.endx = currentHeadRect.x + currentHeadRect.width + RE_DETECTION_MARGIN;
					options.endy = currentHeadRect.y + currentHeadRect.height + RE_DETECTION_MARGIN;
					
					if (options.startx < DETECTION_HORIZ_MARGIN) { options.startx = DETECTION_HORIZ_MARGIN; }
					if (options.starty < DETECTION_VERT_MARGIN) { options.starty = DETECTION_VERT_MARGIN; }
					if (options.endx > webcamBitmap.width - DETECTION_HORIZ_MARGIN) { options.endx = webcamBitmap.width - DETECTION_HORIZ_MARGIN; }
					if (options.endy > webcamBitmap.height - DETECTION_VERT_MARGIN) { options.endy = webcamBitmap.height - DETECTION_VERT_MARGIN; }
				}
				else //use standard size
				{
					options.startx = DETECTION_HORIZ_MARGIN; 
					options.starty = DETECTION_VERT_MARGIN;
					options.endx = webcamBitmap.width - DETECTION_HORIZ_MARGIN;
					options.endy = webcamBitmap.height - DETECTION_VERT_MARGIN;
				}
				
				detector.options = options; //update with new options
				//detector.detect(webcamBitmap);
				detector.detect(webcamBitmap.bitmapData); //perform head trackign
	
				lastDetectionStartTime = new Date().time; //set last detection time to current time
			}
		}
		
		//scales and positions speechballon and head tracking box
		private function renderInfoBasedOnHeadPosition():void
		{
			var scaleRatio:Number = (currentHeadRect.width / 10) / 15;
			scaleRatio = (speechBalloon.scaleX + scaleRatio) / 2;
			faceRectContainer.graphics.clear();
			if (trackingBoxCheckBox.selected)
			{
				faceRectContainer.graphics.lineStyle( 2 );
				faceRectContainer.graphics.drawRect(currentHeadRect.x, currentHeadRect.y, currentHeadRect.width, currentHeadRect.height);
			}
			speechBalloon.scaleX = scaleRatio;
			speechBalloon.scaleY = scaleRatio;
			
			speechBalloon.x = currentHeadRect.x + currentHeadRect.width - (currentHeadRect.width *.15);
			speechBalloon.y = (currentHeadRect.y - speechBalloon.height) + (currentHeadRect.height * .2);
			
		}
		
		//update speech balloon text on every text change
		private function updateSpeechBalloon(evt:Event):void
		{
			speechBalloonTF.text = speechTextArea.text;
		}

	}//end class
	
}//end package
